/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/sdl/graphics.c: Graphics and rendering functionality for
 * SDL-based environments.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/debug.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/sysdep/sdl/internal.h"
#include "src/test.h"  // For is_running_tests().

#ifdef SIL_PLATFORM_WINDOWS
# undef WIN32_LEAN_AND_MEAN  // Work around symbol collision with Mesa gl.h.
# include "src/sysdep/windows/internal.h"
#endif

#ifdef SIL_PLATFORM_MACOSX
# include "src/sysdep/opengl/gl-headers.h"  // See FIXME below.
#endif

#include <SDL.h>
#include <SDL_syswm.h>

#if defined(SIL_PLATFORM_MACOSX) && !SDL_VERSION_ATLEAST(2,0,0)
# error SDL 2.0 or later is required on Mac OS X.
#endif

#if !(defined(SIL_PLATFORM_LINUX) || defined(SIL_PLATFORM_MACOSX))
# define putenv(s) SDL_putenv(s)
#endif

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/**
 * SIL_SDL_CENTER_WINDOW:  If defined, the window created by SDL in
 * windowed mode will be centered on the screen when it is created.  (This
 * doesn't prevent the user from subsequetnly moving the window elsewhere.)
 */
// #define SIL_SDL_CENTER_WINDOW

/**
 * SIL_SDL_DYNAMIC_OPENGL:  If defined, the system's OpenGL library will
 * be loaded when graphics_init() is called.  This is undefined by default
 * to avoid problems on systems without dynamic library support; systems
 * that wish to enable dynamic loading should define this symbol in the
 * system's configuration header.
 */
// #define SIL_SDL_DYNAMIC_OPENGL

#ifdef SIL_SDL_DYNAMIC_OPENGL
# include "src/sysdep/sdl/dyngl.h"
#else
/* Assume all optional features are available. */
# define dyngl_has_compressed_textures()  1
# define dyngl_has_vertex_buffers()       1
# define dyngl_has_shaders()              1
# define dyngl_has_framebuffers()         1
# define dyngl_has_opengl3()              1
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Graphics capability structure.  The display size list is generated at
 * initialization time; the list below is only a default in case SDL
 * doesn't give us any modes. */

static const SysGraphicsSizeEntry default_display_sizes[] = {
    { 640, 480},
    { 800, 480},
    { 800, 600},
    {1024, 768},
    {1280, 720},
    {1280, 800},
    {1600,1200},
    {1920,1080},
    {1920,1200},
};

static SysGraphicsInfo graphics_info = {
    .has_windowed_mode = 1,
    .num_sizes = lenof(default_display_sizes),
    .sizes = default_display_sizes,
};

/*-----------------------------------------------------------------------*/

/* Blacklist for graphics cards known to have no or poor support for
 * certain features despite advertising compatibility.  The vendor and
 * renderer strings are front-anchored pattern matches using simple
 * filename globs (wildcards * and ? supported). */

#define BLACKLIST_NO_FRAMEBUFFERS  (1<<0)  // Disable framebuffers.
#define BLACKLIST_NO_SHADERS       (1<<1)  // Disable shaders.

static const struct {
    const char *vendor;    // Pattern for matching GL_VENDOR.
    const char *renderer;  // Pattern for matching GL_RENDERER.
    int flags;             // BLACKLIST_* flags for this hardware.
} hardware_blacklist[] = {
    // For example:
    // {"NVIDIA Corporation", "GeForce 6600*", BLACKLIST_NO_FRAMEBUFFERS},
    {"NOMATCH", "NOMATCH", 0},  // Dummy entry to avoid compiler warnings.
#ifdef INCLUDE_TESTS
    {"__MATCH__", "__T?ST__", BLACKLIST_NO_FRAMEBUFFERS},
    {"__T?ST__", "__MATCH__", BLACKLIST_NO_SHADERS},
    {"*", "__R*G?X__", BLACKLIST_NO_FRAMEBUFFERS | BLACKLIST_NO_SHADERS},
#endif
};

/*-----------------------------------------------------------------------*/

/* Have we been initialized? */
static uint8_t initted;

#ifdef SIL_SDL_DYNAMIC_OPENGL
/* Have the OpenGL function pointers been set up? */
static uint8_t initted_dyngl;
#endif

/* Has the OpenGL subsystem been initialized? */
static uint8_t initted_opengl;

/* Should we save OpenGL state when closing and reopening the window? */
static uint8_t do_save_opengl_state;

/* Is OpenGL state saved from a previously open window? */
static uint8_t saved_opengl_state;

/* Should we show the mouse pointer? */
static uint8_t show_mouse_pointer;

/* Have we successfully opened a window? */
static uint8_t window_open;

/* Is the current window in fullscreen mode? */
static uint8_t window_fullscreen;

/* Is fullscreen mode selected (for the next mode change)? */
static uint8_t fullscreen;

/* Is multisampling enabled? */
static uint8_t multisample;

/* Is vertical sync enabled? */
static uint8_t vsync;

/* Current display size. */
static int display_width, display_height;

/* Requested depth and stencil sizes. */
static int depth_bits;
static int stencil_bits;

/* Requested number of samples for multisampling. */
static int samples;

/* SDL object for the open window, if any. */
#if SDL_VERSION_ATLEAST(2,0,0)
static SDL_Window *current_window;
static SDL_GLContext window_glcontext;
#else
static SDL_Surface *window_surface;
#endif

#if SDL_VERSION_ATLEAST(2,0,0)
/* Title to use with new windows. */
static char *window_title;
#endif

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * setup_opengl:  Initialize the OpenGL subsystem.  This must be called
 * after a display surface has been created.
 *
 * [Parameters]
 *     width, height: Display size, in pixels.
 */
static int setup_opengl(int width, int height);

/**
 * blacklist_match:  Pattern-matching routine for blacklist checking.
 *
 * [Parameters]
 *     string: String to check.
 *     pattern: Pattern to match against.
 * [Return value]
 *     True if the string matches the pattern, false if not.
 */
static int blacklist_match(const char *string, const char *pattern);

/*************************************************************************/
/***************** Interface: Basic graphics operations ******************/
/*************************************************************************/

const SysGraphicsInfo *sys_graphics_init(void)
{
    PRECOND(!initted, return NULL);

#if !SDL_VERSION_ATLEAST(2,0,0) && defined(SIL_SDL_CENTER_WINDOW)
    /* Center the window in windowed mode. */
    if (setenv("SDL_VIDEO_CENTERED", "1", 1) != 0) {
        DLOG("setenv(SDL_VIDEO_CENTERED=1) failed: %s", SDL_GetError());
    }
#endif

    if (UNLIKELY(SDL_InitSubSystem(SDL_INIT_VIDEO) != 0)) {
        DLOG("SDL video init failed: %s", SDL_GetError());
        goto error_return;
    }

#ifdef DEBUG
# if SDL_VERSION_ATLEAST(2,0,0)
    const char *namebuf = SDL_GetCurrentVideoDriver();
# else
    char namebuf[100];
    SDL_VideoDriverName(namebuf, sizeof(namebuf));
# endif
    DLOG("SDL video driver: %s", namebuf);
#endif

#ifdef SIL_SDL_DYNAMIC_OPENGL
    if (UNLIKELY(SDL_GL_LoadLibrary(NULL) != 0)) {
        DLOG("SDL_GL_LoadLibrary() failed: %s", SDL_GetError());
        goto error_quit_video;
    }
#endif

#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_DisplayMode *modes = NULL;
    int num_modes = SDL_GetNumDisplayModes(0);
    int num_sizes = 0;
    if (num_modes > 0) {
        modes = mem_alloc(sizeof(*modes) * num_modes, 0, MEM_ALLOC_TEMP);
        if (UNLIKELY(!modes)) {
            DLOG("No memory for %d video modes", num_modes);
            num_modes = 0;
        }
        for (int i = 0; i < num_modes; i++) {
            SDL_DisplayMode *mode = &modes[num_sizes];
            int reject = 0;
            int bpp;
            uint32_t dummy;
            if (UNLIKELY(SDL_GetDisplayMode(0, i, mode) != 0)) {
                DLOG("Failed to get display mode %d: %s", i, SDL_GetError());
                reject = 1;
            } else if (!SDL_PixelFormatEnumToMasks(
                           mode->format,
                           &bpp, &dummy, &dummy, &dummy, &dummy)) {
                reject = 1;
            } else if (bpp < 24) {
                reject = 1;
            }
            if (!reject) {
                num_sizes++;
            }
        }
    }
#else
    SDL_Rect **modes = SDL_ListModes(NULL, SDL_FULLSCREEN | SDL_OPENGL);
    /* We should never get -1 if we're requesting fullscreen modes. */
    ASSERT(modes != (SDL_Rect **)-1, modes = NULL);
    int num_sizes = 0;
    if (LIKELY(modes)) {
        for (; modes[num_sizes] != NULL; num_sizes++) {
            /* Just count the modes here. */
        }
    }
#endif
    if (UNLIKELY(!modes) || UNLIKELY(!num_sizes)) {
        DLOG("SDL returned no%s display modes!", modes ? " valid" : "");
        graphics_info.num_sizes = lenof(default_display_sizes);
        graphics_info.sizes = default_display_sizes;
    } else {
        SysGraphicsSizeEntry *sizes =
            mem_alloc(sizeof(*graphics_info.sizes) * num_sizes, 0, 0);
        if (UNLIKELY(!sizes)) {
            DLOG("No memory for display size list");
#if SDL_VERSION_ATLEAST(2,0,0)
            mem_free(modes);
#endif
            goto error_quit_video;
        }
        for (int i = 0; i < num_sizes; i++) {
#if SDL_VERSION_ATLEAST(2,0,0)
            sizes[i].width  = modes[i].w;
            sizes[i].height = modes[i].h;
#else
            sizes[i].width  = modes[i]->w;
            sizes[i].height = modes[i]->h;
#endif
        }
#ifdef SIL_PLATFORM_LINUX
        /* Some versions of X.org and/or display drivers seem to report
         * all known resolutions as "supported", when in fact some of
         * those resolutions are unavailable and will cause fatal BadValue
         * errors if used.  Find the actual display dimensions and filter
         * out any larger modes. */
        int x11_width = 0, x11_height = 0;
# if SDL_VERSION_ATLEAST(2,0,0)
        SDL_DisplayMode mode;
        if (UNLIKELY(SDL_GetDesktopDisplayMode(0, &mode) != 0)) {
            DLOG("Failed to get display mode info: %s", SDL_GetError());
        } else {
            x11_width = mode.w;
            x11_height = mode.h;
        }
# else
        SDL_SysWMinfo info;
        SDL_VERSION(&info.version);
        if (UNLIKELY(!SDL_GetWMInfo(&info))) {
            DLOG("Failed to get display pointer: %s", SDL_GetError());
        } else {
            (*info.info.x11.lock_func)();
            Display *display = info.info.x11.display;
            x11_width = DisplayWidth(display, DefaultScreen(display));
            x11_height = DisplayHeight(display, DefaultScreen(display));
            (*info.info.x11.unlock_func)();
        }
# endif
        for (int i = 0; i < num_sizes; i++) {
            if ((x11_width > 0 && sizes[i].width > x11_width)
             || (x11_height > 0 && sizes[i].height > x11_height)) {
                num_sizes--;
                memmove(&sizes[i], &sizes[i+1],
                        sizeof(*sizes) * (num_sizes - i));
                i--;
            }
        }
#endif
        graphics_info.num_sizes = num_sizes;
        graphics_info.sizes = sizes;
    }
#if SDL_VERSION_ATLEAST(2,0,0)
    mem_free(modes);
#endif

    do_save_opengl_state = 1;
    window_open = 0;
    fullscreen = 0;
    multisample = 0;
    vsync = 1;
    depth_bits = 16;
    stencil_bits = 0;
    samples = 1;
#if SDL_VERSION_ATLEAST(2,0,0)
    current_window = NULL;
    window_title = NULL;
#else
    window_surface = NULL;
#endif

    initted = 1;
    initted_dyngl = 0;
    initted_opengl = 0;
    return &graphics_info;

  error_quit_video:
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
  error_return:
    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_cleanup(void)
{
    PRECOND(initted, return);

    if (graphics_info.sizes != default_display_sizes) {
        mem_free((void *)graphics_info.sizes);
        graphics_info.num_sizes = lenof(default_display_sizes);
        graphics_info.sizes = default_display_sizes;
    }

    if (initted_opengl) {
        opengl_cleanup();
        initted_opengl = 0;
        saved_opengl_state = 0;
    }

#if SDL_VERSION_ATLEAST(2,0,0)
    if (window_open) {
        SDL_GL_DeleteContext(window_glcontext);
        SDL_DestroyWindow(current_window);
        current_window = NULL;
    }
    mem_free(window_title);
    window_title = NULL;
#endif

    SDL_QuitSubSystem(SDL_INIT_VIDEO);

#ifdef SIL_SDL_DYNAMIC_OPENGL
    if (initted_dyngl) {
        dyngl_cleanup();
        initted_dyngl = 0;
    }
#endif

    window_open = 0;
    initted = 0;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_device_width(void)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_DisplayMode mode = {.w = 0};
    ASSERT(SDL_GetCurrentDisplayMode(0, &mode) == 0);
    return mode.w;
#else
    return SDL_GetVideoInfo()->current_w;
#endif
}

/*-----------------------------------------------------------------------*/

int sys_graphics_device_height(void)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_DisplayMode mode = {.h = 0};
    ASSERT(SDL_GetCurrentDisplayMode(0, &mode) == 0);
    return mode.h;
#else
    return SDL_GetVideoInfo()->current_h;
#endif
}

/*-----------------------------------------------------------------------*/

int sys_graphics_set_display_attr(const char *name, va_list args)
{
    if (strcmp(name, "depth_bits") == 0) {
        const int value = va_arg(args, int);
        if (value < 0) {
            DLOG("Invalid value for attribute depth_bits: %d", value);
            return 0;
        }
        depth_bits = value;
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, depth_bits);
        opengl_set_depth_bits(depth_bits);
        return 1;
    }

    if (strcmp(name, "multisample") == 0) {
        const int value = va_arg(args, int);
        if (value <= 0) {
            DLOG("Invalid value for attribute multisample: %d", value);
            return 0;
        }
        samples = value;
        multisample = (samples > 1);
        SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, multisample);
        SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, samples);
        return 1;
    }

    if (strcmp(name, "preserve_state") == 0) {
        do_save_opengl_state = (va_arg(args, int) != 0);
        return 1;
    }

    if (strcmp(name, "stencil_bits") == 0) {
        const int value = va_arg(args, int);
        if (value < 0) {
            DLOG("Invalid value for attribute stencil_bits: %d", value);
            return 0;
        }
        stencil_bits = value;
        SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, stencil_bits);
        opengl_set_stencil_bits(stencil_bits);
        return 1;
    }

    if (strcmp(name, "vsync") == 0) {
        vsync = (va_arg(args, int) != 0);
#if SDL_VERSION_ATLEAST(2,0,0)
        SDL_GL_SetSwapInterval(vsync);
#else
        SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, vsync);
#endif
        return 1;
    }

    if (strcmp(name, "window") == 0) {
        fullscreen = (va_arg(args, int) == 0);
        return 1;
    }

    return 0;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_set_display_mode(int width, int height)
{
    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

#if SDL_VERSION_ATLEAST(2,0,0)
# ifdef SIL_PLATFORM_MACOSX
    /* On OSX we have to choose between either OpenGL 2.1 (ancient) or
     * OpenGL 3.2 core profile (which requires some hacks in our common
     * OpenGL code).  We choose the latter since drivers for it are
     * more likely to be kept up to date.  Since we require at least
     * OSX 10.7 (which provides OpenGL 3.2), we don't need to worry
     * about falling back to an older version. */
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK,
                        SDL_GL_CONTEXT_PROFILE_CORE);
# endif

    SDL_DisplayMode desktop_mode = {.w = 0, .h = 0};
    ASSERT(SDL_GetDesktopDisplayMode(0, &desktop_mode) == 0);
    uint32_t flags = (SDL_WINDOW_OPENGL | SDL_WINDOW_ALLOW_HIGHDPI
                      | (fullscreen ? SDL_WINDOW_FULLSCREEN : 0));
#else
    const uint32_t flags = SDL_OPENGL | (fullscreen ? SDL_FULLSCREEN : 0);
#endif

#if !SDL_VERSION_ATLEAST(2,0,0)
    /* Check ahead of time if the call is going to fail, to avoid killing
     * an existing window unnecessarily. */
    if (!SDL_VideoModeOK(width, height, 32, flags)) {
        DLOG("Video mode rejected by SDL: %dx%d%s", width, height,
             fullscreen ? ", full-screen" : "");
        return 0;
    }
#endif

#if defined(INCLUDE_TESTS) && !SDL_VERSION_ATLEAST(2,0,0)
    /* If testing and if we're going to open a fullscreen window, preserve
     * the position of the mouse pointer and restore it after the window
     * has been opened.  (Normally, SDL 1.2 repositions the mouse pointer
     * to the center of the screen when setting a fullscreen mode.) */

    float mouse_oldx = -1, mouse_oldy = -1;
    if (is_running_tests() && fullscreen) {
# if defined(SIL_PLATFORM_LINUX)
        SDL_SysWMinfo info;
        SDL_VERSION(&info.version);
        if (UNLIKELY(!SDL_GetWMInfo(&info))) {
            DLOG("Failed to get display pointer: %s", SDL_GetError());
        } else if (info.info.x11.display) {
            (*info.info.x11.lock_func)();
            Display *display = info.info.x11.display;
            Window root = RootWindow(display, DefaultScreen(display));
            Window unused_r, unused_c;
            int x, y, unused_wx, unused_wy;
            unsigned int unused_mask;
            if (UNLIKELY(!XQueryPointer(display, root, &unused_r, &unused_c,
                                        &x, &y, &unused_wx, &unused_wy,
                                        &unused_mask))) {
                DLOG("Failed to get pointer position");
            } else {
                const int w = DisplayWidth(display, DefaultScreen(display));
                const int h = DisplayHeight(display, DefaultScreen(display));
                mouse_oldx = (float)x / (float)w;
                mouse_oldy = (float)y / (float)h;
            }
            (*info.info.x11.unlock_func)();
        } else {
            DLOG("NOTE: can't get pointer position (SDL is not using X11),"
                 " will be reset by fullscreen mode");
        }
# else  // !LINUX
        DLOG("NOTE: can't get pointer position, will be reset by fullscreen"
             " mode");
# endif
    }
#endif  // INCLUDE_TESTS && SDL < 2.0.0

    if (window_open && width == display_width && height == display_height) {
        int window_depth, window_buffers, window_samples, window_stencil;
        if (SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &window_depth) == 0
         && SDL_GL_GetAttribute(SDL_GL_MULTISAMPLEBUFFERS, &window_buffers) == 0
         && SDL_GL_GetAttribute(SDL_GL_MULTISAMPLESAMPLES, &window_samples) == 0
         && SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE, &window_stencil) == 0
         && window_depth >= depth_bits
         && (!multisample || (window_buffers && window_samples >= samples))
         && window_stencil >= stencil_bits) {
            /* Window size and parameters are unchanged -- we're just
             * (possibly) switching into or out of fullscreen mode. */
            if ((fullscreen != 0) == window_fullscreen) {
                /* No change at all! */
                return 1;
            } else {
#if SDL_VERSION_ATLEAST(2,0,0)
                const int new_fullscreen = !window_fullscreen;
                const uint32_t flag =
                    (new_fullscreen ? SDL_WINDOW_FULLSCREEN : 0);
                const int ok =
                    (SDL_SetWindowFullscreen(current_window, flag) == 0);
#else
                const int ok = SDL_WM_ToggleFullScreen(window_surface);
#endif
                if (ok) {
                    window_fullscreen = !window_fullscreen;
                    return 1;
                } else {
                    DLOG("Failed to toggle fullscreen: %s", SDL_GetError());
                    /* Fall through to regular close/open handling. */
                }
            }
        }
    }

    /* If we already had a window open, save OpenGL state to memory because
     * the GL context will be destroyed. */
    if (window_open && initted_opengl && do_save_opengl_state
     && !saved_opengl_state) {
        opengl_save_state();
        saved_opengl_state = 1;
    }

#if SDL_VERSION_ATLEAST(2,0,0)
# ifdef SIL_SDL_CENTER_WINDOW
    const int position = SDL_WINDOWPOS_CENTERED;
# else
    const int position = SDL_WINDOWPOS_UNDEFINED;
# endif
    if (window_open) {
        SDL_GL_DeleteContext(window_glcontext);
        SDL_DestroyWindow(current_window);
        window_open = 0;
    }
    current_window = SDL_CreateWindow(
        window_title, position, position, width, height, flags);
    if (current_window) {
        window_glcontext = SDL_GL_CreateContext(current_window);
        if (!window_glcontext) {
            DLOG("Failed to create GL context, discarding window: %s",
                 SDL_GetError());
            SDL_DestroyWindow(current_window);
            current_window = NULL;
        }
    }
    if (!current_window)
#else  // SDL 1.2
    window_surface = SDL_SetVideoMode(width, height, 32, flags);
    if (!window_surface)
#endif
    {
        DLOG("Failed to set video mode (%dx%d%s): %s", width, height,
             fullscreen ? ", full-screen" : "", SDL_GetError());
        return 0;
    }

#if defined(INCLUDE_TESTS) && !SDL_VERSION_ATLEAST(2,0,0)
    if (is_running_tests() && mouse_oldx >= 0 && mouse_oldy >= 0) {
        SDL_WarpMouse(iroundf(mouse_oldx * width),
                      iroundf(mouse_oldy * height));
    }
#endif

    /* Hide the mouse pointer if requested. */
    SDL_ShowCursor(show_mouse_pointer ? SDL_ENABLE : SDL_DISABLE);

    /* In SDL 2.x, we have to explicitly pump and flush the input queue to
     * get the window's focus flag set and avoid breaking in strange ways
     * (for example, without this code, SDL 2.0.2 on OSX will randomly
     * crash with a CALayer exception). */
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_PumpEvents();
    SDL_Event sdl_event;
    while (SDL_PollEvent(&sdl_event)) { /*nothing*/ }
#endif

#ifdef SIL_SDL_DYNAMIC_OPENGL
    /* Initialize OpenGL function pointers if not yet done.  We can't do
     * this any earlier because on Windows, we only get the OpenGL 1.1(!)
     * subset of functions if no context has been created. */
    if (!initted_dyngl) {
        if (!dyngl_init()) {
            return 0;
        }
        initted_dyngl = 1;
    }
#endif

    /* Check that we got the OpenGL buffer parameters we requested. */
    int window_depth, window_buffers, window_samples, window_stencil;
    // FIXME: SDL_GL_GetAttribute(SDL_GL_{DEPTH,STENCIL}_BITS) is broken in SDL
    // for core profiles: https://bugzilla.libsdl.org/show_bug.cgi?id=2060
#ifdef SIL_PLATFORM_MACOSX
    GLint value;
    opengl_clear_error();
    /* For totally incomprehensible reasons, the second parameter has to
     * be GL_DEPTH and not GL_DEPTH_ATTACHMENT when using the default
     * framebuffer. */
    glGetFramebufferAttachmentParameteriv(
        GL_DRAW_FRAMEBUFFER, GL_DEPTH,
        GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE, &value);
    if (value != GL_NONE) {
        glGetFramebufferAttachmentParameteriv(
            GL_FRAMEBUFFER, GL_DEPTH,
            GL_FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE, &value);
        window_depth = value;
    } else {
        window_depth = 0;
    }
    glGetIntegerv(GL_SAMPLE_BUFFERS, &value);
    window_buffers = value;
    glGetIntegerv(GL_SAMPLES, &value);
    window_samples = value;
    glGetFramebufferAttachmentParameteriv(
        GL_DRAW_FRAMEBUFFER, GL_STENCIL,
        GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE, &value);
    if (value != GL_NONE) {
        glGetFramebufferAttachmentParameteriv(
            GL_FRAMEBUFFER, GL_STENCIL,
            GL_FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE, &value);
        window_stencil = value;
    } else {
        window_stencil = 0;
    }
    const GLenum gl_error = glGetError();
    if (UNLIKELY(gl_error != GL_NO_ERROR)) {
        DLOG("Failed to get OpenGL attributes: 0x%04X", gl_error);
        return 0;
    }
#else
    if (UNLIKELY(SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &window_depth) != 0)
     || UNLIKELY(SDL_GL_GetAttribute(SDL_GL_MULTISAMPLEBUFFERS,
                                     &window_buffers) != 0)
     || UNLIKELY(SDL_GL_GetAttribute(SDL_GL_MULTISAMPLESAMPLES,
                                     &window_samples) != 0)
     || UNLIKELY(SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE,
                                     &window_stencil) != 0))
    {
        DLOG("Failed to get OpenGL attributes: %s", SDL_GetError());
        return 0;
    }
#endif
    if (window_depth < depth_bits) {
        DLOG("Requested %d depth bits but only got %d",
             depth_bits, window_depth);
        return 0;
    }
    if (multisample) {
        if (!window_buffers) {
            DLOG("Requested multisampling but didn't get it");
            return 0;
        }
        if (window_samples < samples) {
            DLOG("Requested multisampling with %d samples but only got %d",
                 samples, window_samples);
            return 0;
        }
    }
    if (window_stencil < stencil_bits) {
        DLOG("Requested %d stencil bits but only got %d",
             stencil_bits, window_stencil);
        return 0;
    }

    /* Set up OpenGL (if not yet done) or restore saved state now that we
     * have a context to work with. */
    if (!initted_opengl) {
        if (UNLIKELY(!setup_opengl(width, height))) {
            return 0;
        }
        initted_opengl = 1;
    } else if (saved_opengl_state) {
        opengl_restore_state();
        saved_opengl_state = 0;
    }

    /* Enable multisampling if requested. */
    if (multisample) {
        glEnable(GL_MULTISAMPLE);
    } else {
        glDisable(GL_MULTISAMPLE);
    }

    display_width = width;
    display_height = height;
    window_open = 1;
    window_fullscreen = (fullscreen != 0);
    return 1;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_display_is_window(void)
{
    return window_open && !window_fullscreen;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_set_window_title(const char *title)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    mem_free(window_title);
    window_title = mem_strdup(title, 0);
    if (UNLIKELY(!window_title)) {
        DLOG("No memory to copy window title");
    }
    if (window_open) {
        SDL_SetWindowTitle(current_window, title);
    }
#else
    SDL_WM_SetCaption(title, title);
#endif
}

/*-----------------------------------------------------------------------*/

void sys_graphics_set_window_icon(SysTexture *texture)
{
    const int width = sys_texture_width(texture);
    const int height = sys_texture_height(texture);
#ifdef SIL_PLATFORM_WINDOWS
    if (width != 32 || height != 32) {
        DLOG("Invalid texture size for icon (must be 32x32)");
        return;
    }
#endif

    uint8_t *pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                       0, 0, width, height);
    if (!pixels) {
        DLOG("Failed to lock texture for icon");
        return;
    }

    SDL_Surface *surface = SDL_CreateRGBSurfaceFrom(
        pixels, width, height, 32, width*4,
#ifdef IS_LITTLE_ENDIAN
        0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000
#else
        0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF
#endif
    );
    if (UNLIKELY(!surface)) {
        DLOG("Failed to create surface");
        sys_texture_unlock(texture, 0);
        return;
    }
#if SDL_VERSION_ATLEAST(2,0,0)
    if (window_open) {
        SDL_SetWindowIcon(current_window, surface);
    }
#else
    SDL_WM_SetIcon(surface, NULL);
#endif
    SDL_FreeSurface(surface);
    sys_texture_unlock(texture, 0);
}

/*-----------------------------------------------------------------------*/

void sys_graphics_show_mouse_pointer(int on)
{
    show_mouse_pointer = (on != 0);
    SDL_ShowCursor(show_mouse_pointer ? SDL_ENABLE : SDL_DISABLE);
}

/*-----------------------------------------------------------------------*/

int sys_graphics_get_mouse_pointer_state(void)
{
    return show_mouse_pointer;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_get_frame_period(int *numerator_ret, int *denominator_ret)
{
    /* Set fallback values in case of error. */
    *numerator_ret = 0;
    *denominator_ret = 1;

#if SDL_VERSION_ATLEAST(2,0,0)
    int display;
    if (window_open) {
        display = SDL_GetWindowDisplayIndex(current_window);
        if (UNLIKELY(display < 0)) {
            DLOG("Failed to get window's display index: %s", SDL_GetError());
            return;
        }
    } else {
        display = 0;
    }

    SDL_DisplayMode mode;
    if (UNLIKELY(SDL_GetCurrentDisplayMode(display, &mode) != 0)) {
        DLOG("SDL_GetCurrentDisplayMode(%d) failed: %s", display,
             SDL_GetError());
        return;
    }
    if (mode.refresh_rate > 0) {
        *numerator_ret = 1;
        *denominator_ret = mode.refresh_rate;
    }
#else
    /* SDL 1.2 doesn't provide a way to get the frame rate of a video mode. */
#endif
}

/*-----------------------------------------------------------------------*/

int sys_graphics_has_focus(void)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    return window_open
        && (SDL_GetWindowFlags(current_window) & SDL_WINDOW_INPUT_FOCUS);
#else
    return (SDL_GetAppState() & SDL_APPINPUTFOCUS) != 0;
#endif
}

/*-----------------------------------------------------------------------*/

void sys_graphics_start_frame(int *width_ret, int *height_ret)
{
    *width_ret = display_width;
    *height_ret = display_height;

#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_PROCESS_END);
    debug_record_cpu_phase(DEBUG_CPU_GPU_WAIT_START);
    opengl_sync();
    debug_record_cpu_phase(DEBUG_CPU_GPU_WAIT_END);
#endif

#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_RENDER_START);
#endif
    opengl_start_frame();
}

/*-----------------------------------------------------------------------*/

void sys_graphics_finish_frame(void)
{
#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_RENDER_END);
#endif
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_GL_SwapWindow(current_window);
#else
    SDL_GL_SwapBuffers();
#endif
#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_PROCESS_START);
#endif
}

/*-----------------------------------------------------------------------*/

void sys_graphics_sync(int flush)
{
    opengl_sync();
    if (flush) {
        opengl_free_dead_resources(1);
    }
}

/*************************************************************************/
/******************** SDL-internal exported routines *********************/
/*************************************************************************/

struct SDL_Window *sdl_window(void)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    return window_open ? current_window : NULL;
#else
    return NULL;
#endif
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static int setup_opengl(int width, int height)
{
    /* Set up feature flags for the generic OpenGL module, disabling
     * anything known not to work on this particular hardware. */
    int gl_flags = 0;
#ifndef SIL_PLATFORM_MACOSX
    gl_flags |= OPENGL_FEATURE_NATIVE_QUADS;
#endif
    if (dyngl_has_shaders()) {
        gl_flags |= OPENGL_FEATURE_SHADERS;
    }
    if (dyngl_has_framebuffers()) {
        gl_flags |= OPENGL_FEATURE_FRAMEBUFFERS;
#ifndef SIL_PLATFORM_MACOSX
        gl_flags |= OPENGL_FEATURE_GENERATEMIPMAP;
#endif
    }
    if (dyngl_has_vertex_buffers()) {
        gl_flags |= OPENGL_FEATURE_FAST_STATIC_VBO
                 |  OPENGL_FEATURE_FAST_DYNAMIC_VBO;
    }
#ifdef SIL_PLATFORM_MACOSX
    gl_flags |= OPENGL_FEATURE_MANDATORY_VAO;
#endif

    const char *gl_vendor = (const char *)glGetString(GL_VENDOR);
    const char *gl_renderer = (const char *)glGetString(GL_RENDERER);
    for (int i = 0; i < lenof(hardware_blacklist); i++) {
        if (blacklist_match(gl_vendor, hardware_blacklist[i].vendor)
         && blacklist_match(gl_renderer, hardware_blacklist[i].renderer)) {
            if ((hardware_blacklist[i].flags & BLACKLIST_NO_FRAMEBUFFERS)
             && (gl_flags & OPENGL_FEATURE_FRAMEBUFFERS)) {
                DLOG("Disabling framebuffers on this hardware due to poor"
                     " performance or other problems.  (Blacklist match:"
                     " vendor=[%s] renderer=[%s])",
                     hardware_blacklist[i].vendor,
                     hardware_blacklist[i].renderer);
                gl_flags &= ~OPENGL_FEATURE_FRAMEBUFFERS;
                gl_flags &= ~OPENGL_FEATURE_GENERATEMIPMAP;
            }
            if ((hardware_blacklist[i].flags & BLACKLIST_NO_SHADERS)
             && (gl_flags & OPENGL_FEATURE_SHADERS)) {
                DLOG("Disabling shaders on this hardware due to poor"
                     " performance or other problems.  (Blacklist match:"
                     " vendor=[%s] renderer=[%s])",
                     hardware_blacklist[i].vendor,
                     hardware_blacklist[i].renderer);
                gl_flags &= ~OPENGL_FEATURE_SHADERS;
            }
            break;
        }
    }

    /* Initialize the generic OpenGL module. */
    if (UNLIKELY(!opengl_init(width, height, gl_flags))) {
        DLOG("Failed to set up OpenGL!");
        return 0;
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

static int blacklist_match(const char *string, const char *pattern)
{
    PRECOND(string != NULL, return 0);
    PRECOND(pattern != NULL, return 0);

    while (*string) {
        if (*pattern == '*') {
            pattern += strspn(pattern, "*"); // Extra asterisks have no effect.
            if (!*pattern) {
                /* A trailing asterisk matches anything. */
                return 1;
            } else {
                /* Recursively try to match the remaining part of the
                 * pattern at each position. */
                for (; *string; string++) {
                    if (blacklist_match(string, pattern)) {
                        return 1;
                    }
                }
                return 0;  // No matches found.
            }
        } else {
            if (*pattern != '?' && *string != *pattern) {
                return 0;
            }
            pattern++;
            string++;
        }
    }

    return pattern[strspn(pattern,"*")] == 0;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
