/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/opengl/shader.c: Tests for OpenGL shader handling.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/sysdep/opengl/shader-table.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Number of calls made to generate_shader_source(). */
static int shader_source_calls;

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * generate_shader_source:  Shader source generator for testing error
 * paths.  Behavior is controlled by the fog and alpha_test parameters:
 *
 * fog == 0, alpha_test == 0: Causes a compile error in the vertex shader.
 * fog == 0, alpha_test == 1: Causes a compile error in the fragment shader.
 * fog == 1, alpha_test == 0: Causes a link error.
 * fog == 1, alpha_test == 1: Genreates empty source code for both shaders.
 *
 * Setting color_uniform nonzero will cause the function to return an error
 * without generating any source.
 */
static int generate_shader_source(
    UNUSED GLuint vertex_shader, UNUSED GLuint fragment_shader,
    UNUSED GLenum primitive_type, UNUSED int position_count,
    UNUSED int texcoord_count, UNUSED GLenum texcolor_type,
    UNUSED int tex_offset, UNUSED int color_count, UNUSED int color_uniform,
    int fog, int alpha_test, UNUSED GLenum alpha_comparison)
{
    #ifdef SIL_OPENGL_SHADER_VERSION
    # define SIL_OPENGL_SHADER_VERSION_STRING _STR1(SIL_OPENGL_SHADER_VERSION)
    # define _STR1(x) _STR2(x)
    # define _STR2(x) #x
    # define SHADER_VERSION_LINE_ "#version "SIL_OPENGL_SHADER_VERSION_STRING"\n",
    #else
    # define SHADER_VERSION_LINE_ /*nothing*/
    #endif

    #if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
    # define VERTEX_IN    "in"
    # define GL_FRAGCOLOR "frag_color"
    #else
    # define VERTEX_IN    "attribute"
    # define GL_FRAGCOLOR "gl_FragColor"
    #endif

    #ifdef SIL_OPENGL_ES
    # define HIGHP "highp"
    # define LOWP "lowp"
    #else
    # define HIGHP ""
    # define LOWP ""
    #endif

    static const char *vs_source[] = {
        SHADER_VERSION_LINE_
        VERTEX_IN" "HIGHP" vec4 position;\n",
        "void main() {\n",
        "    gl_Position = position;\n",
        "}\n",
    };
    static const char *fs_source[] = {
        SHADER_VERSION_LINE_
#if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
        "out "LOWP" vec4 frag_color;\n",
#endif
        "void main() {\n",
        "    "GL_FRAGCOLOR" = vec4(1.0, 1.0, 1.0, 1.0);\n",
        "}\n",
    };

    shader_source_calls++;

    if (color_uniform) {
        return 0;

    } else if (!fog && !alpha_test) {
        static const char *bad_vs_source[] = {
            SHADER_VERSION_LINE_
            VERTEX_IN" "HIGHP" vec4 position;\n",
            "void main() {\n",
            "    gl_Position = pos_ition;\n",
            "}\n",
        };
        glShaderSource(vertex_shader, lenof(bad_vs_source), bad_vs_source,
                       NULL);
        glShaderSource(fragment_shader, lenof(fs_source), fs_source, NULL);

    } else if (!fog && alpha_test) {
        static const char *bad_fs_source[] = {
            SHADER_VERSION_LINE_
#if defined(SIL_OPENGL_SHADER_VERSION) && SIL_OPENGL_SHADER_VERSION >= 130
            "out "LOWP" vec4 frag_color;\n",
#endif
            "void main() {\n",
            "    "GL_FRAGCOLOR" = v_ec4(1.0, 1.0, 1.0, 1.0);\n",
            "}\n",
        };
        glShaderSource(vertex_shader, lenof(vs_source), vs_source, NULL);
        glShaderSource(fragment_shader, lenof(bad_fs_source), bad_fs_source,
                        NULL);

    } else if (fog && !alpha_test) {
        static const char *bad_fs_source[] = {
            "void ma_in() {\n",
            "    "GL_FRAGCOLOR" = vec4(1.0, 1.0, 1.0, 1.0);\n",
            "}\n",
        };
        glShaderSource(vertex_shader, lenof(vs_source), vs_source, NULL);
        glShaderSource(fragment_shader, lenof(bad_fs_source), bad_fs_source,
                       NULL);

    } else {
        glShaderSource(vertex_shader, 0, vs_source, NULL);
        glShaderSource(fragment_shader, 0, fs_source, NULL);

    }

    return 1;

    #undef HIGHP
}

/*-----------------------------------------------------------------------*/

/**
 * generate_shader_key:  Shader key generator for testing error paths.
 * Setting color_count > 0 will result in an error; all other parameter
 * combinations are accepted.
 */
static uint32_t generate_shader_key(
    UNUSED GLenum primitive_type, UNUSED int position_count,
    UNUSED int texcoord_count, UNUSED GLenum texcolor_type,
    UNUSED int tex_offset, int color_count, UNUSED int color_uniform,
    int fog, int alpha_test, UNUSED GLenum alpha_comparison)
{
    if (color_count > 0) {
        return INVALID_SHADER_KEY;
    } else {
        return (color_uniform * 2 + fog) * 2 + alpha_test;
    }
}

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

static int wrap_test_opengl_shader(void);
static int do_test_opengl_shader(void);
int test_opengl_shader(void)
{
    return run_tests_in_window(wrap_test_opengl_shader);
}

static int wrap_test_opengl_shader(void)
{
    if (!opengl_has_features(OPENGL_FEATURE_SHADERS)) {
        SKIP("Shaders not available.");
    }
    return do_test_opengl_shader();
}

DEFINE_GENERIC_TEST_RUNNER(do_test_opengl_shader)

TEST_INIT(init)
{
    opengl_set_shader_callbacks(NULL, NULL, 0, 0);
    CHECK_TRUE(shader_table_init(0, 1));
    graphics_start_frame();
    return 1;
}

TEST_CLEANUP(cleanup)
{
    graphics_finish_frame();
    opengl_shader_clear_all();
    opengl_set_shader_callbacks(NULL, NULL, 0, 0);
    CHECK_TRUE(shader_table_init(0, 1));
    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

#undef FAIL_ACTION
#define FAIL_ACTION  failed = 1

TEST(test_shader_key)
{
    static const int texcolors[] = {
        TEXCOLOR_RGBA, TEXCOLOR_A,
#if defined(SIL_OPENGL_ES) && defined(GL_OES_EGL_image_external)
        TEXCOLOR_EXTERNAL,
#endif
    };
    static const GLenum comparisons[] = {
        GL_LESS, GL_LEQUAL, GL_GEQUAL, GL_GREATER};

    int failed = 0;

    sys_texture_apply(NULL);   // So we don't have to worry about saving it.

    /**** Nested test loops. ****/

    for (int is_point = 0; is_point <= 1; is_point++) {
     for (int position_count = 2; position_count <= 4; position_count++) {
      for (int i_texcolor = 0; i_texcolor < 1+lenof(texcolors); i_texcolor++) {
       for (int use_tex_offset = 0; use_tex_offset <= 1; use_tex_offset++) {
        for (int use_vertex_color = 0; use_vertex_color <= 1; use_vertex_color++) {
         for (int use_fixed_color = 0; use_fixed_color <= 1; use_fixed_color++) {
          for (int use_fog = 0; use_fog <= 1; use_fog++) {
           for (int alpha_test = 0; alpha_test < 1+lenof(comparisons); alpha_test++) {

                /**** Single test implementation. ****/

                /* Create dummy objects with just the information needed by
                 * the the shader routines. */
                SysPrimitive primitive;
                SysTexture texture;
                primitive.type = is_point ? GL_POINTS : GL_TRIANGLES;
                primitive.position_size = position_count;
                primitive.color_size = use_vertex_color ? 4 : 0;
                if (i_texcolor > 0) {
                    primitive.texcoord_size = 2;
                    texture.color_type = texcolors[i_texcolor - 1];
                } else {
                    primitive.texcoord_size = 0;
                    texture.color_type = TEXCOLOR_RGBA;
                }

                /* Check how many shaders are stored in the table. */
                const int shaders_loaded = shader_table_used();

                /* Select the shader (which should create it). */
                ASSERT(opengl_current_texture == NULL);
                opengl_current_texture = &texture;
                if (opengl_select_shader(
                        &primitive, use_tex_offset, use_fixed_color, use_fog,
                        alpha_test != 0,
                        alpha_test ? comparisons[alpha_test-1] : 0) <= 0) {
                    FAIL("Did not select a new shader for"
                         " (%u,%u,%u,%u,%u,%u,%u,%u)",
                         is_point, position_count, i_texcolor,
                         use_tex_offset, use_vertex_color,
                         use_fixed_color, use_fog, alpha_test);
                    failed = 1;
                }
                opengl_current_texture = NULL;

                /* Check that a new shader was in fact created. */
                if (shader_table_used() != shaders_loaded+1) {
                    FAIL("New shader was not created for"
                         " (%u,%u,%u,%u,%u,%u,%u,%u)",
                         is_point, position_count, i_texcolor,
                         use_tex_offset, use_vertex_color,
                         use_fixed_color, use_fog, alpha_test);
                    failed = 1;
                }

           }  // use_alpha_test
          }  // use_fog
         }  // use_fixed_color
        }  // use_vertex_color
       }  // use_tex_offset
      }  // i_texcolor
     }  // position_count
    }  // is_point

    return !failed;
}

#undef FAIL_ACTION
#define FAIL_ACTION  return 0

/*-----------------------------------------------------------------------*/

TEST(test_select_shader_memory_failure)
{
    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    CHECK_TRUE(shader_table_init(0, 1));

    TEST_mem_fail_after(0, 1);
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0);
    TEST_mem_fail_after(-1, 0);
    CHECK_INTEQUAL(shader_table_used(), 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_select_shader_unsupported_params)
{
    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    sys_texture_apply(NULL);  // So we don't have to worry about saving it.
    SysTexture texture;
    texture.color_type = TEXCOLOR_RGBA;

    primitive.position_size = 5;
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0);
    primitive.position_size = 3;

    primitive.texcoord_size = 5;
    opengl_current_texture = &texture;
    /* Make sure opengl_current_texture is cleared on test failure. */
    CHECK_TRUE((opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0)
               || (opengl_current_texture = NULL));
    opengl_current_texture = NULL;
    primitive.texcoord_size = 0;

    primitive.color_size = 5;
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0);
    primitive.color_size = 0;

    /* Invalid alpha comparison. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 1, 0) < 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_set_uniform_with_no_shader)
{
    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 5;  // Invalid value.
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0);

    /* Just make sure these don't crash. */
    opengl_set_uniform_int(UNIFORM_TEXTURE, 0);
    opengl_set_uniform_float(UNIFORM_ALPHA_REF, 0);
    opengl_set_uniform_vec2(UNIFORM_TEX_OFFSET, &(Vector2f){0,0});
    opengl_set_uniform_vec4(UNIFORM_FIXED_COLOR, &(Vector4f){0,0,0,0});
    opengl_set_uniform_mat4(UNIFORM_TRANSFORM, &mat4_identity);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_shader_table_fixed_size)
{
    /* Cycle through all 6 permutations of the 3 position count values,
     * to ensure that the oldest shader is always evicted first. */
    static const uint8_t position_counts[6][3] = {
        {2,3,4},
        {2,4,3},
        {3,2,4},
        {3,4,2},
        {4,2,3},
        {4,3,2},
    };

    for (int test = 0; test < lenof(position_counts); test++) {
        if (!shader_table_init(3, 0)) {
            FAIL("shader_table_init() failed for test %u", test);
        }
        opengl_shader_table_overflow_count = 0;

        SysPrimitive primitive;
        primitive.type = GL_TRIANGLES;
        primitive.color_size = 0;
        primitive.texcoord_size = 0;

        /* First generate shaders for each position count to fill the table. */
        for (int i = 0; i < 3; i++) {
            primitive.position_size = position_counts[test][i];
            if (opengl_select_shader(&primitive, 0, 0, 0, 0, 0) <= 0) {
                FAIL("Did not select a new shader for test %u index %d",
                     test, i);
            }
        }
        if (shader_table_used() != 3) {
            FAIL("Did not store 3 shaders for test %u", test);
        }

        /* Generate a fourth shader, and check that the table size remains
         * unchanged and an entry is evicted. */
        primitive.color_size = 4;
        if (opengl_select_shader(&primitive, 0, 0, 0, 0, 0) <= 0) {
            FAIL("Did not select a new shader for test %u", test);
        }
        if (shader_table_used() != 3) {
            FAIL("Stored shader count increased for test %u", test);
        }
        if (opengl_shader_table_overflow_count != 1) {
            FAIL("Did not record a shader table eviction for test %u", test);
        }

        /* Check that the last two shaders in the test set were not evicted. */
        primitive.color_size = 0;
        for (int i = 1; i < 3; i++) {
            primitive.position_size = position_counts[test][i];
            if (opengl_select_shader(&primitive, 0, 0, 0, 0, 0) <= 0) {
                FAIL("Did not select a new shader for test %u index %d",
                     test, i);
            }
            if (shader_table_used() != 3) {
                FAIL("Stored shader count increased for test %u index %d",
                     test, i);
            }
            if (opengl_shader_table_overflow_count != 1) {
                FAIL("Test %u index %d was incorrectly evicted", test, i);
            }
        }

        /* Check that the first shader in the test set was evicted. */
        primitive.position_size = position_counts[test][0];
        if (opengl_select_shader(&primitive, 0, 0, 0, 0, 0) <= 0) {
            FAIL("Did not select a new shader for test %u index 0", test);
        }
        if (shader_table_used() != 3) {
            FAIL("Stored shader count increased for test %u index 0", test);
        }
        if (opengl_shader_table_overflow_count != 2) {
            FAIL("Test %u index 0 should have been evicted but was not", test);
        }

        opengl_shader_clear_all();
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_shader_table_resize_memory_failure)
{
    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    CHECK_MEMORY_FAILURES(shader_table_init(1, 1));

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 1);

    TEST_mem_fail_after(0, 1);
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 1, 0, 0, 0) > 0);
    TEST_mem_fail_after(-1, 0);
    CHECK_INTEQUAL(shader_table_used(), 1);  // Expect overwrite.

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_shader_table_clear)
{
    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    CHECK_TRUE(shader_table_init(0, 1));

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 1);

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 1, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 2);

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 2);

    opengl_shader_clear_all();
    CHECK_INTEQUAL(shader_table_used(), 0);

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_shader_table_clear_fixed_size)
{
    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    CHECK_TRUE(shader_table_init(5, 0));

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 1);

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 1, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 2);

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 2);

    opengl_shader_clear_all();
    CHECK_INTEQUAL(shader_table_used(), 0);

    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) > 0);
    CHECK_INTEQUAL(shader_table_used(), 1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_shader_custom_source)
{
    opengl_set_shader_callbacks(generate_shader_source, generate_shader_key,
                                5, 0);

    SysPrimitive primitive;
    primitive.type = GL_TRIANGLES;
    primitive.position_size = 3;
    primitive.texcoord_size = 0;
    primitive.color_size = 0;

    shader_source_calls = 0;

    /* Key generation fails. */
    primitive.color_size = 4;
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 1, 1, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 0);
    CHECK_INTEQUAL(shader_source_calls, 0);
    primitive.color_size = 0;

    /* Shader source generation fails. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 1, 0, 0, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 1);
    CHECK_INTEQUAL(shader_source_calls, 1);

    /* Vertex shader compilation fails. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 2);
    CHECK_INTEQUAL(shader_source_calls, 2);

    /* Fragment shader compilation fails. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 1, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 3);
    CHECK_INTEQUAL(shader_source_calls, 3);

    /* Shader program link fails. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 1, 0, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 4);
    CHECK_INTEQUAL(shader_source_calls, 4);

    /* Empty source for both shaders. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 1, 1, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 5);
    CHECK_INTEQUAL(shader_source_calls, 5);

    /* Subsequent calls on failed parameter sets should not trigger
     * source generation again. */
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 1, 0, 0, 0) < 0);
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 0, 0) < 0);
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 0, 1, 0) < 0);
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 1, 0, 0) < 0);
    CHECK_TRUE(opengl_select_shader(&primitive, 0, 0, 1, 1, 0) < 0);
    CHECK_INTEQUAL(shader_table_used(), 5);
    CHECK_INTEQUAL(shader_source_calls, 5);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
